from yta_general_utils.programming.validator.number import NumberValidator
from datetime import datetime


def get_rfc_datetime(year: int = 1900, month: int = 1, day: int = 1, hour: int = 0, minute: int = 0):
    """
    Receives a date based on provided parameters and turns it into
    a RFC datetime, returning it.

    TODO: Explain this better, please
    """
    dt = datetime(year, month, day, hour, minute, 0).isoformat() + 'Z'

    return dt

def seconds_to_hh_mm_ss(seconds: int):
    """
    Turn the provided amount of 'seconds' to a time
    in the format HH:MM:SS.
    """
    if not NumberValidator.is_positive_number(seconds, do_include_zero = True):
        raise Exception('The provided amount of "seconds" is not a positive number (including zero).')
    
    hh = seconds // 3600
    mm = (seconds % 3600) // 60 
    ss = seconds % 60

    return f'{hh:02}:{mm:02}:{ss:02}'

# TODO: Make this more configurable
def current_datetime() -> str:
    """
    Return the current time moment, as a string, in
    the 'DD/MM/YYYY HH:MM:SS' format.
    """
    return datetime.now().strftime("%d/%m/%Y %H:%M:%S")

# TODO: Build more methods to simplify our work



from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.number import NumberValidator
from yta_general_utils.programming.validator.error_message import ErrorMessage
from typing import Union


class ParameterValidator:
    """
    Class to wrap and simplify the method parameters
    validation, so you are able to validate if a
    parameter is a mandatory string, if it is a 
    positive value, and normalize the exception
    messages.

    The type of the parameters passed to each method
    validation is not validated, so pay attention to
    what you pass there.

    Each method will raise an Exception with a custom
    message if failing.
    """

    @staticmethod
    def validate_mandatory(
        name: str,
        value: any
    ) -> None:
        """
        Validate if the provided 'value' is not None.
        """
        if value is None:
            raise Exception(ErrorMessage.parameter_not_provided(name))

    @staticmethod
    def validate_string(
        name: str,
        value: Union[str, None],
        do_accept_empty: bool = True
    ) -> None:
        """
        Validate if the provided 'value' is of string type.
        """
        if (
            (
                value is not None and
                not PythonValidator.is_string(value)
            ) or
            (
                value == '' and
                not do_accept_empty
            )
        ):
            raise Exception(ErrorMessage.parameter_is_not_string(name))
        
    @staticmethod
    def validate_mandatory_string(
        name: str,
        value: str,
        do_accept_empty: bool = True
    ) -> None:
        """
        Validate if the provided 'value' is a non-empty
        string.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_string(name, value)

        if (
            not do_accept_empty and
            value == ''
        ):
            raise Exception(ErrorMessage.parameter_not_provided(value))
    
    @staticmethod
    def validate_bool(
        name: str,
        value: Union[bool, None]
    ) -> None:
        """
        Validate if the provided 'value' is bool value or not.
        """
        if (
            value is not None and
            not PythonValidator.is_boolean(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_boolean(name))
        
    @staticmethod
    def validate_mandatory_bool(
        name: str,
        value: bool
    ) -> None:
        """
        Validate if the provided 'value' is not None and is
        a boolean value.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_bool(name, value)

    @staticmethod
    def validate_dict(
        name: str,
        value: Union[dict, None]
    ) -> None:
        """
        Validate if the provided 'value' is a dict.
        """
        if (
            value is not None and
            not PythonValidator.is_dict(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_dict(name))

    @staticmethod
    def validate_mandatory_dict(
        name: str,
        value: dict
    ) -> None:
        """
        Validate if the provided 'value' is not None and is
        a dict.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_dict(name, value)

    @staticmethod
    def validate_callable(
        name: str,
        value: Union[callable, None]
    ) -> None:
        """
        Validate if the provided 'value' is callable.
        """
        if (
            value is not None and
            not PythonValidator.is_callable(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_callable(name))
        
    @staticmethod
    def validate_mandatory_callable(
        name: str,
        value: callable
    ) -> None:
        """
        Validate if the provided 'value' is not None and
        is callable.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_callable(name, value)

    @staticmethod
    def validate_int(
        name: str,
        value: int
    ) -> None:
        """
        Validate if the provided 'value' is an int.
        """
        if (
            value is not None and
            not NumberValidator.is_int(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_int(name))

    @staticmethod
    def validate_mandatory_int(
        name: str,
        value: int
    ) -> None:
        """
        Validate if the provided 'value' is not None
        and is an int.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_int(name, value)

    def validate_positive_int(
        name: str,
        value: int,
        do_include_zero: bool = True
    ) -> None:
        """
        Validate if the provided 'value' is a positive
        int number.
        """
        if (
            value is not None and
            (
                not NumberValidator.is_int(value) or
                not NumberValidator.is_positive_number(value, do_include_zero = do_include_zero)
            )
        ):
            raise Exception(ErrorMessage.parameter_is_not_positive_int(name))
        
    def validate_mandatory_positive_int(
        name: str,
        value: int,
        do_include_zero: bool = True
    ):
        """
        Validate if the provided 'value' is not None
        and is a positive int number.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_positive_int(name, value, do_include_zero = do_include_zero)

    @staticmethod
    def validate_float(
        name: str,
        value: float
    ) -> None:
        """
        Validate if the provided 'value' is a float.
        """
        if (
            value is not None and
            not NumberValidator.is_float(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_float(name))

    @staticmethod
    def validate_mandatory_float(
        name: str,
        value: float
    ) -> None:
        """
        Validate if the provided 'value' is not None
        and is a float.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_float(name, value)

    @staticmethod
    def validate_positive_float(
        name: str,
        value: float,
        do_include_zero: bool = True
    ):
        """
        Validate if the provided 'value' is a positive
        int number.
        """
        if (
            value is not None and
            (
                not NumberValidator.is_float(value) or
                not NumberValidator.is_positive_number(value, do_include_zero = do_include_zero)
            )
        ):
            raise Exception(ErrorMessage.parameter_is_not_positive_float(name))
        
    def validate_mandatory_positive_float(
        name: str,
        value: float,
        do_include_zero: bool = True
    ):
        """
        Validate if the provided 'value' is not None
        and is a positive float number.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_positive_float(name, value, do_include_zero = do_include_zero)


    @staticmethod
    def validate_positive_number(
        name: str,
        value: Union[float, int, None],
        do_include_zero: bool = True
    ) -> None:
        """
        Validate if the provided 'value' is a positive number
        or not.
        """
        if (
            value is not None and
            not NumberValidator.is_positive_number(value, do_include_zero = do_include_zero)
        ):
            raise Exception(ErrorMessage.parameter_is_not_positive_number(name))
        
    @staticmethod
    def validate_mandatory_positive_number(
        name: str,
        value: Union[float, int],
        do_include_zero: bool = True
    ) -> None:
        """
        Validate if the provided 'value' is not None and is a
        positive number or not.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_positive_number(name, value, do_include_zero = do_include_zero)
        
    @staticmethod
    def validate_instance_of(
        name: str,
        value: Union[object, None],
        cls: Union[type, str, list[Union[type, str]]]
    ) -> None:
        """
        Validate if the provided 'value' is an instance of the
        given 'cls' class or classes.
        """
        if (
            value is not None and
            not PythonValidator.is_instance(value, cls)
        ):
            raise Exception(ErrorMessage.parameter_is_not_instance_of(name, cls))

    @staticmethod
    def validate_mandatory_instance_of(
        name: str,
        value: object,
        cls: list[Union[object, str]]
    ) -> None:
        """
        Validate if the provided 'value' is not None and is an
        instance of the given 'cls' classes.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_instance_of(name, value, cls)

    @staticmethod
    def validate_class_of(
        name: str,
        value: Union[type, None],
        cls: list[Union[type, str]]
    ) -> None:
        """
        Validate if the provided 'value' is one of the given
        'cls' classes.
        """
        if (
            value is not None and
            not PythonValidator.is_class(value, cls)
        ):
            raise Exception(ErrorMessage.parameter_is_not_class_of(name, cls))
        
    @staticmethod
    def validate_mandatory_class_of(
        name: str,
        value: type,
        cls: list[Union[type, str]]
    ) -> None:
        """
        Validate if the provided 'value' is not None and is 
        one of the given 'cls' classes.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_class_of(name, value, cls)

    @staticmethod
    def validate_list_of_numbers(
        name: str,
        value: Union[list[Union[int, float]], None]
    ) -> None:
        """
        Validate if the provided 'value' is a list of numbers.
        """
        if (
            value is not None and
            not PythonValidator.is_list_of_numbers(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_list_of_numbers(name))

    @staticmethod
    def validate_mandatory_list_of_numbers(
        name: str,
        value: list[Union[int, float]]
    ) -> None:
        """
        Validate if the provided 'value' is not None and is a
        list of numbers.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_list_of_numbers(name, value)

    @staticmethod
    def validate_list_of_positive_numbers(
        name: str,
        value: Union[list[Union[int, float]], None],
        do_include_zero: bool = True
    ) -> None:
        """
        Validate if the provided 'value' is a list of 
        positive numbers.
        """
        if (
            value is not None and
            not PythonValidator.is_list_of_positive_numbers(value, do_include_zero = do_include_zero)
        ):
            raise Exception(ErrorMessage.parameter_is_not_list_of_positive_numbers(name))

    @staticmethod
    def validate_mandatory_list_of_positive_numbers(
        name: str,
        value: list[Union[int, float]],
        do_include_zero: bool = True
    ) -> None:
        """
        Validate if the provided 'value' is not None and is a
        list of positive numbers.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_list_of_positive_numbers(name, value, do_include_zero = do_include_zero)

    @staticmethod
    def validate_list_of_string(
        name: str,
        value: Union[list[str], None]
    ) -> None:
        """
        Validate if the provided 'value' is a list of string
        values.
        """
        if (
            value is not None and
            not PythonValidator.is_list_of_string(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_list_of_string(name))
        
    @staticmethod
    def validate_mandatory_list_of_string(
        name: str,
        value: list[str]
    ) -> None:
        """
        Validate if the provided 'value' is not None and is
        a list of string values.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_list_of_string(name, value)

    @staticmethod
    def validate_list_of_instances(
        name: str,
        value: Union[list[object], None]
    ) -> None:
        """
        Validate if the provided 'value' is a list of 
        instances.
        """
        if (
            value is not None and
            not PythonValidator.is_list_of_instances(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_list_of_instances(name))
        
    @staticmethod
    def validate_mandatory_list_of_instances(
        name: str,
        value: list[object]
    ) -> None:
        """
        Validate if the provided 'value' is not None and
        is a list of instances.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_list_of_instances(name, value)
        
    @staticmethod
    def validate_list_of_these_instances(
        name: str,
        value: Union[list[object], None],
        cls: Union[list[Union[type, str]], str, type]
    ) -> None:
        """
        Validate if the provided 'value' is a list of 
        instances of the given 'cls' class or classes.
        """
        if (
            value is not None and
            not PythonValidator.is_list_of_these_instances(value, cls)
        ):
            raise Exception(ErrorMessage.parameter_is_not_list_of_these_instances(name, cls))
        
    @staticmethod
    def validate_mandatory_list_of_these_instances(
        name: str,
        value: list[object],
        cls: Union[list[Union[type, str]], str, type]
    ) -> None:
        """
        Validate if the provided 'value' is not None
        and is a list of instances of the given 'cls'
        class or classes.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_list_of_these_instances(name, value, cls)

    @staticmethod
    def validate_list_of_classes(
        name: str,
        value: Union[list[type], None]
    ) -> None:
        """
        Validate if the provided 'value' is a list of
        classes.
        """
        if (
            value is not None and
            not PythonValidator.is_list_of_classes(value)
        ):
            raise Exception(ErrorMessage.parameter_is_not_list_of_classes(name))
        
    @staticmethod
    def validate_mandatory_list_of_classes(
        name: str,
        value: list[type]
    ) -> None:
        """
        Validate if the provided 'value' is not None
        and is a list of classes.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_list_of_classes(name, value)

    @staticmethod
    def validate_list_of_these_classes(
        name: str,
        value: Union[list[type], None],
        cls: Union[list[Union[type, str]], str, type]
    ) -> None:
        """
        Validate if the provided 'value' is a list
        of the given 'cls' class or classes.
        """
        if (
            value is not None and
            not PythonValidator.is_list_of_these_classes(value, cls)
        ):
            raise Exception(ErrorMessage.parameter_is_not_list_of_these_classes(name, cls))

    @staticmethod
    def validate_mandatory_list_of_these_classes(
        name: str,
        value: list[type],
        cls: Union[list[Union[type, str]], str, type]
    ) -> None:
        """
        Validate if the provided 'value' is not None
        and is a list of the given 'cls' class or
        classes.
        """
        ParameterValidator.validate_mandatory(name, value)
        ParameterValidator.validate_list_of_these_classes(name, value, cls)

    @staticmethod
    def validate_dict_has_keys(
        name: str,
        value: dict,
        keys: list[str]
    ) -> None:
        """
        Validate if the provided 'value' dict has or
        not the also given 'keys', raising an Exception
        if not.
        """
        missing_keys = PythonValidator.get_missing_keys_in_dict(value, keys)

        if missing_keys.length > 0:
            raise Exception(ErrorMessage.keys_are_missing_in_dict(name, missing_keys))


from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from requests import Response
from typing import Union

import json


class FileEncoding(Enum):
    """
    The different file encoding we accept. This Enum
    has been created to be used as encoding when
    writing on a file.

    # TODO: Maybe rename to 'TextEncoding'? Maybe use
    # an external library or move this toanother
    # module (not library) (?)
    """

    UTF8 = 'utf8'

class FileWriter:
    """
    Class to simplify and encapsulate the functionality related with
    writting files.
    """

    @staticmethod
    def write_binary_file(
        binary_data: bytes,
        output_filename: str
    ) -> str:
        """
        Writes the provided 'binary_data' in the 'filename'
        file. It replaces the previous content if existing.

        This method returns the filename as written.
        """
        ParameterValidator.validate_mandatory('binary_data', binary_data)
        ParameterValidator.validate_mandatory_string('output_filename', output_filename, do_accept_empty = False)

        return _write_binary(
            binary_data,
            output_filename
        )

    @staticmethod
    def write_json_to_file(
        dict: dict,
        output_filename: str
    ) -> str:
        """
        Writes the provided 'dict' as a json into the 'filename'.

        This method returns the filename as written.

        @param
            **dict**
            Python dictionary that will be stored as a json.

            **output_filename**
            File path in which we are going to store the information.
        """
        ParameterValidator.validate_mandatory_dict('dict', dict)
        ParameterValidator.validate_mandatory_string('output_filename', output_filename, do_accept_empty = False)
        
        return FileWriter.write_file(
            text = json.dumps(dict, indent = 4),
            output_filename = output_filename
        )

    @staticmethod
    def write_file(
        text: str,
        output_filename: str,
        encoding: Union[str, FileEncoding, None] = FileEncoding.UTF8
    ) -> str:
        """
        Writes the provided 'text' in the 'filename' file. It
        replaces the previous content if existing.

        This method returns the filename that has been
        written.
        """
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = True)
        ParameterValidator.validate_mandatory_string('output_filename', output_filename, do_accept_empty = False)

        encoding = (
            None
            if encoding == None else
            FileEncoding.to_enum(encoding).value
        )

        return _write(text, encoding, output_filename)

    @staticmethod
    def write_file_by_chunks_from_response(
        response: Response,
        output_filename: str
    ) -> str:
        """
        Iterates over the provided 'response' and writes its content
        chunk by chunk in the also provided 'output_filename'.

        TODO: If you find a better way to handle this you are free to
        create new methods and move them into a new file.

        This method returns the filename that has been
        written.
        """
        ParameterValidator.validate_mandatory_instance_of('response', response, Response)
        ParameterValidator.validate_mandatory_string('output_filename', output_filename, do_accept_empty = False)
        
        CHUNK_SIZE = 32768

        # TODO: Make this method work with a common Iterator parameter
        # and not an specific response, please
        with open(output_filename, 'wb') as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return output_filename
    
def _write(
    content: str,
    encoding: Union[str, FileEncoding, None],
    output_filename: str
) -> str:
    """
    Returns the 'output_filename' that has been
    written.
    """
    encoding = (
        None
        if encoding == None else
        FileEncoding.to_enum(encoding).value
    )

    with open(
        file = output_filename,
        mode = 'w',
        encoding = encoding
    ) as f:
        f.write(content)

    return output_filename

def _write_binary(
    content: bytes,
    output_filename: str
) -> str:
    """
    Returns the 'output_filename' that has been
    written.
    """
    with open(
        file = output_filename,
        mode = 'wb'
    ) as f:
        f.write(content)

    return output_filename


from yta_general_utils.file.enums import SubtitleFileExtension
from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.file.reader import FileReader
from yta_general_utils.subtitles.dataclasses import SubtitleLine, Subtitles
from datetime import timedelta

import xml.etree.ElementTree as ET
import re
import json


class SubtitlesParser:
    """
    Class to simplify the way we parse subtitles.
    """

    @staticmethod
    def parse_from_filename(
        filename: str
    ):
        """
        Parse the subtitles file with the provided
        'filename'.
        """
        extension = SubtitleFileExtension.to_enum(get_file_extension(filename))
        content = FileReader.read(filename)

        return SubtitlesParser.parse_from_content(content, extension)

    @staticmethod
    def parse_from_content(
        text: str,
        extension: SubtitleFileExtension
    ):
        """
        Parse the provided 'text' content of a subtitles
        file with the given 'extension' format.
        """
        extension = SubtitleFileExtension.to_enum(extension)

        parse_fn = {
            SubtitleFileExtension.SRV1: lambda text: parse_srv1(text),
            SubtitleFileExtension.SRV2: lambda text: parse_srv2(text),
            SubtitleFileExtension.SRV3: lambda text: parse_srv3(text),
            SubtitleFileExtension.JSON3: lambda text: parse_json3(json.loads(text)),
            SubtitleFileExtension.VTT: lambda text: parse_vtt(text),
            SubtitleFileExtension.TTML: lambda text: parse_ttml(text)
        }.get(extension, None)

        return parse_fn(text) if parse_fn else None
    

def parse_srv1(text: str) -> Subtitles:
    """
    Parse the content of a SRV1 subtitles file.

    A srv1 file looks like this:
    <transcript>
    <text start="2.906" dur="3">TEXT 1</text>
    <text start="7.907" dur="3.914">TEXT 2</text>
    """
    subtitles = [
        SubtitleLine(
            text = text_element.text.strip() if text_element.text else '',
            start_time = round(float(text_element.get('start', 0)), 3),
            duration = round(float(text_element.get('dur', 0)), 3)
        ) for text_element in ET.fromstring(text).findall('.//text')
    ]

    return Subtitles(subtitles)

def parse_srv2(text: str) -> Subtitles:
    """
    Parse the content of a SRV2 subtitles file.

    A srv2 file looks like this:
    <transcript>
    <text t="2.906" d="3">TEXT 1</text>
    <text t="7.907" d="3.914">TEXT 2</text>
    """
    subtitles = [
        SubtitleLine(
            text = text_element.text.strip() if text_element.text else '',
            start_time = int(text_element.get('t', 0)),
            duration = int(text_element.get('d', 0))
        ) for text_element in ET.fromstring(text).findall('.//text')
    ]

    return Subtitles(subtitles)

def parse_srv3(text: str) -> Subtitles:
    """
    Parse the content of a SRV3 subtitles file.

    A srv3 file looks like this:
    TODO: Put example
    """
    subtitles = []
    for paragraph in ET.fromstring(text).findall('.//p'):
        paragraph_start_time = int(paragraph.attrib['t'])
        duration = int(paragraph.attrib['d'])
        
        for word in paragraph.findall('s'):
            # Words with no 't' are the first ones and
            # its start_time is the paragraph start_time
            # and the ones with 't', that 't' is the 
            # 'start_time' relative to the paragraph
            relative_word_start_time = int(word.attrib['t']) if 't' in word.attrib else 0
            text = word.text

            subtitles.append(
                SubtitleLine(
                    text = text,
                    start_time = paragraph_start_time + relative_word_start_time,
                    # This duration will be changed it later in
                    # post-processing based on the next word
                    # 'start_time'
                    duration = duration
                )
            )

    for index, subtitle in enumerate(subtitles[1:], start = 1):
        subtitles[index - 1].duration = subtitle.start_time - subtitles[index - 1].start_time

    return Subtitles(subtitles)

def parse_json3(json: dict) -> Subtitles:
    """
    Parse the content of a JSON3 subtitles file.

    A json3 file looks like this:
    TODO: Put example
    """
    subtitles = []
    for paragraph in json.get('events', []):
        paragraph_start_time = paragraph.get('tStartMs', 0)
        paragraph_duration = paragraph.get('dDurationMs', 0)
        for word in paragraph.get('segs', []):
            relative_word_start_time = int(word['tOffsetMs']) if 'tOffsetMs' in word else 0
            text = word['utf8']
            # We drop any '\n' word as we don't care
            if text == '\n':
                continue

            subtitles.append(
                SubtitleLine(
                    text = text,
                    start_time = paragraph_start_time + relative_word_start_time,
                    # This duration will be changed it later in
                    # post-processing based on the next word
                    # 'start_time'
                    duration = paragraph_duration
                )
            )

    for index, subtitle in enumerate(subtitles[1:], start = 1):
        subtitles[index - 1].duration = subtitle.start_time - subtitles[index - 1].start_time

    return Subtitles(subtitles)

def parse_vtt(text: str) -> Subtitles:
    """
    Parse the content of a VTT subtitles file.

    A vtt file looks like this:
    TODO: Put example
    """
    # TODO: We need to fix an error with the processing
    # that ends generating an empty line followed by a
    # line without blank spaces. Both of them should not
    # be there, but they are (see Notion for more info)
    def clean_text_tags(text: str):
        return re.sub(r'<[^>]+>', '', text).strip()
        #return re.sub(r'<.*?>', '', text).strip()
    
    # We remove the first 3 header lines
    text = '\n'.join(text.splitlines()[3:])
    
    # Remove comments, empty lines or headers and a
    # text we don't want nor need
    text = text.strip()
    text = text.replace(' align:start position:0%', '')

    # Split the content in blocks according to time line
    subtitle_blocks = re.split(r'\n(?=\d{2}:\d{2}:\d{2}\.\d{3})', text)
    
    subtitles = []
    for block in subtitle_blocks:
        lines = block.splitlines()

        # First row is time
        time_range = lines[0].split(' --> ')
        start_time_str = time_range[0].strip()
        end_time_str = time_range[1].strip()

        start_time = _time_to_ms(start_time_str)
        end_time = _time_to_ms(end_time_str)
        duration = end_time - start_time
        
        # Next rows are the text
        text_segments = [
            line.strip()
            for line in lines[1:]
        ]

        # Clean tags and unify
        full_text = ' '.join([clean_text_tags(text) for text in text_segments])

        # The sentences with duration 10 are the ones 
        # that contain the new sentence that is said
        # in the video, so those texts are the ones we
        # must keep for our purpose.
        if duration == 10:
            subtitles[-1] = SubtitleLine(full_text, subtitles[-1].start_time, subtitles[-1].duration + 10)
        else:
            subtitles.append(SubtitleLine(full_text, start_time, duration))
    
    return Subtitles(subtitles)

def parse_ttml(text: str) -> Subtitles:
    # This is the namespace used in this kind of file
    namespace = {'tt': 'http://www.w3.org/ns/ttml'}

    # Look for paragraphs containing it
    subtitles = []
    for p in ET.fromstring(text).findall('.//tt:body//tt:div//tt:p', namespace):
        start_time_str = p.get('begin')
        end_time_str = p.get('end')
        
        if start_time_str and end_time_str:
            start_time = _time_to_ms(start_time_str)
            end_time = _time_to_ms(end_time_str)
            duration = end_time - start_time

            text = ' '.join(p.itertext()).strip()

            # Guardamos el subtítulo como un objeto
            subtitles.append(SubtitleLine(text, start_time, duration))
    
    return Subtitles(subtitles)

# TODO: This should be in a 'date' helper but
# this is very specific so we can't reuse it,
# we need a more general method
def _time_to_ms(time_str):
    """
    Transform a HH:MM:SS:ssss date string to a time
    in milliseconds.
    """
    return int(timedelta(
        hours = int(time_str[:2]),
        minutes = int(time_str[3:5]),
        seconds = int(time_str[6:8]),
        milliseconds = int(time_str[9:])
    ).total_seconds() * 1000)


from yta_general_utils.file.writer import FileWriter
from yta_general_utils.dataclasses import FileReturn
from typing import Union

import requests


def get_file(
    url: str,
    output_filename: Union[str, None] = None
):
    """
    This method sends a request to the provided 'url'
    if provided and obtains the file content (if 
    possible). It will write the obtained file locally
    as 'output_filename' if provided.

    This method returns the file content data as 
    obtained from the requested (.content field).
    """
    content = requests.get(url).content

    if output_filename:
        FileWriter.write_binary_file(content, output_filename)

    return content

def get_file_by_chunks(
    url: str,
    output_filename: Union[str, None] = None
):
    """
    Download the file from the given 'url' chunk
    by chunk to be able to handle big files.

    This method returns the file content data as 
    obtained from the requested (.content field).
    """
    CHUNK_SIZE = 8192

    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        content = b''.join(response.iter_content(CHUNK_SIZE))

    # content = b''
    # with requests.get(url, stream = True) as response:
    #     response.raise_for_status()

    #     #with open(output_filename, 'wb') if output_filename else nullcontext() as file:
    #         for chunk in response.iter_content(CHUNK_SIZE):
    #             content += chunk
    #             # if file:
    #             #     file.write(chunk)

    if output_filename:
        FileWriter.write_binary_file(content, output_filename)

    return content

def download_file(
    url: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Receives a downloadable url as 'url' and downloads that
    file in our system as 'output_filename'.
    """
    if output_filename is None:
        raise Exception('No "output_filename" provided.')
    
    file = get_file(url, output_filename)
    
    return FileReturn(
        file,
        # TODO: What about this? I don't know the
        # expected extension nor the type...
        None,
        output_filename
    )


from yta_general_utils.programming.path import get_project_abspath
from dotenv import load_dotenv

import os


class Environment:
    """
    Class to handle the environment of the project in
    which you are executing this code perfectly.
    """

    @staticmethod
    def load_current_project_dotenv():
        """
        Load the current project environment '.env' configuration
        file. The current project is the one in which the code
        is being executed (the code in which you call this method,
        not the library in which it is written).

        Any project in which you are importing this library, the
        '.env' file on its main folder will be loaded.
        """
        load_dotenv(os.path.join(get_project_abspath(), '.env'))

    @staticmethod
    def get_current_project_env(
        variable: str
    ):
        """
        Load the current project environment '.env' configuration
        file and get the value of the 'variable' if existing.

        This method makes a 'load_dotenv' call within the current
        project absolute path any time you call it, so it ensures
        the value is correctly loaded if available.

        You don't need to do 'load_dotenv()' to call this method,
        it will do for you =).
        """
        Environment.load_current_project_dotenv()

        return os.getenv(variable)



"""
In this file we handle Tiktok url and we parse
them to obtain basic information and check if
they are valid ones or not.

We have two different types of valid urls:
- Long: 'https://www.tiktok.com/@ahorayasabesque/video/7327001175616703777?\_t=8jqq93LWqsC&\_r=1'
- Short: 'https://vm.tiktok.com/ZGeSJ6YRA'
"""
from yta_general_utils.programming.regular_expressions import GeneralRegularExpression
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.web.tiktok.dataclasses import TiktokUrl

import requests


class TiktokUrlParser:
    """
    Class to simplify the way we parse Tiktok
    videos urls.
    """

    @staticmethod
    def is_valid(url: str):
        """
        Check if the provided Tiktok video 'url' is
        valid or not. A valid Tiktok url can be a 
        short url or a long url.
        """
        if not PythonValidator.is_string(url):
            raise Exception('The provided "url" parameter is not a string.')
        
        return _is_short_tiktok_url(url) or _is_long_tiktok_url(url)
    
    @staticmethod
    def parse(
        url: str
    ) -> TiktokUrl:
        """
        Parse the provided 'url' and return a TiktokUrl
        dataclass instance containing the author username,
        the video id and the long-format url, or raises
        an Exception if the given 'url' is not valid.
        """
        if not PythonValidator.is_string(url):
            raise Exception('The provided "url" parameter is not a string.')

        if not TiktokUrlParser.is_valid(url):
            raise Exception('The provided "url" is not a valid tiktok video url.')
        
        url = _clean(url)
        if not _is_long_tiktok_url(url):
            url = _short_tiktok_url_to_long_tiktok_url(url)

        aux = url.split('/')

        return TiktokUrl(
            username = aux[len(aux) - 3],
            video_id = aux[len(aux) - 1],
            url = url
        )
    
def _is_short_tiktok_url(url: str):
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')
    
    return GeneralRegularExpression.TIKTOK_SHORT_VIDEO_URL.parse(url)

def _is_long_tiktok_url(url: str):
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')
    
    return GeneralRegularExpression.TIKTOK_LONG_VIDEO_URL.parse(url)

def _short_tiktok_url_to_long_tiktok_url(url: str):
    """
    Transforms the provided short tiktok 'url' to 
    the long format and returns it.
    """
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')
    
    if not _is_short_tiktok_url(url):
        raise Exception('No "url" provided is not a short tiktok url.')

    return requests.get(url).url

def _clean(url: str):
    """
    Removes any additional parameter that is after a
    question mark sign.
    """
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')
    
    url = url.split('?')[0] if '?' in url else url

    return url


"""
Module to validate parameters, values, etc.

Some interesting information below:
The 'inspect' module (native in python) is very interesting because
it has some methods to check if a user defined function or things 
like that.

'f.__qualname__' is very interesting thing as it works
with the full path, not as 'f.__name__' does, this is
one example of it value when doing it from a local file:

'Example.example.__qualname__'
test_discord_video.<locals>.Example.example

And yes, the 'example' is a @staticmethod defined in the
Example class that is contained in the 'test_discord_video'
file
"""
from enum import Enum
from array import array
from typing import Union

import numpy as np
import inspect
import re
import validators


class PythonValidator:
    """
    Class to simplify and encapsulate the functionality related with
    parameters and variables validation.

    This class has been created to simplify the way it work and 
    replace the old ParameterValidator that was using too many 
    different methods being imported and generating cyclic import
    issues.

    We have some equivalent methods that do not need to pass the class
    as a parameter, so we can only work with the class name and avoid
    imports that can cause issues.
    """

    @staticmethod
    def is_instance(
        element: object,
        cls: Union[str, type, list]
    ) -> bool:
        """
        Check if the provided 'element' is an instance of the provided
        class (or classes) 'cls'. An instance is not the same as a
        class. The 'cls' parameter can be the class or the string name
        of that class, or a list of them (even mixed).

        This method is useful if you want to check if it belongs to a
        class but without importing the class and passing it as a 
        parameter to avoid cyclic import issues.
        """
        # TODO: Please, simplify this code below but
        # paying attention to avoid infinite loops by
        # reusing other methods of this class

        # 1. Single value and it is a class
        if (
            not PythonValidator.is_list(cls) and
            isinstance(cls, type)
        ):
            return isinstance(element, cls)

        # 2. Single value and it is a string
        if (
            not PythonValidator.is_list(cls) and
            isinstance(cls, str)
        ):
            return getattr(type(element), '__name__', None) == cls
            
        # 3. Multiple values and they are all classes
        if (
            PythonValidator.is_list(cls) and
            all(
                isinstance(cls_item, type)
                for cls_item in cls
            )
        ):
            return any(
                isinstance(element, cls_item)
                for cls_item in cls
            )

        # 4. Multiple values and they are all strings
        if (
            PythonValidator.is_list(cls) and
            all(
                isinstance(cls_item, str)
                for cls_item in cls
            )
        ):
            return getattr(type(element), '__name__', None) in cls

        # 5. Multiple values and they are mixed classes and strings
        if (
            PythonValidator.is_list(cls) and
            all(
                (
                    isinstance(cls_item, str) or
                    isinstance(cls_item, type)
                )
                for cls_item in cls
            )
        ):
            element_class_name = getattr(type(element), '__name__', None)

            cls_str_list = [
                cls_item
                for cls_item in cls
                if isinstance(cls_item, str)
            ]
            cls_classes_list = [
                cls_item
                for cls_item in cls
                if isinstance(cls_item, type)
            ]

            return any(
                (
                    isinstance(element, cls_item)
                    for cls_item in cls_classes_list
                ) or
                (
                    element_class_name in cls_str_list
                )
            )
        
        # TODO: Maybe some parameter provided is invalid
        return False
    
    @staticmethod
    def is_an_instance(
        element: object
    ) -> bool:
        """
        Check if the provided 'element' is an instance of any class.
        """
        return PythonValidator.is_instance(element, object)
        # This below is just another alternative I want to keep
        return getattr(element, '__name__', None) is None
    
    @staticmethod
    def is_class(
        element,
        cls: Union[list[Union[type, str]], str, type]
    ) -> bool:
        """
        Check if the provided 'element' is the provided class 'cls'.
        A class is not the same as an instance of that class. The
        'classes' parameter can be the class or the string name of that
        class. 

        You can pass the string name of the class to avoid import in 
        the file where you are calling this method.
        """
        # TODO: Check this method, I'm not sure that it
        # is working properly         
        if (
            not PythonValidator.is_string(cls) and
            not PythonValidator.is_a_class(cls) and
            not PythonValidator.is_list_of_string(cls) and
            not PythonValidator.is_list_of_classes(cls)
        ):
            raise Exception(f'The provided "cls" parameter "{str(cls)}" is not a string nor a class.')

        # To list
        cls = (
            [cls]
            if not PythonValidator.is_list(cls) else
            cls
        )

        # To str
        cls = [
            (
                class_item
                if PythonValidator.is_string(class_item) else
                class_item.__name__ # the name of the class
            ) for class_item in cls
        ]
        
        if PythonValidator.is_string(cls):
            if cls == 'str':
                return PythonValidator.is_string(element)

            # TODO: This is not working at all, I need to check why
            # and to apply more ifs (maybe)
            return getattr(element, '__name__', None) is cls
        
        return (
            PythonValidator.is_a_class(element) and
            element.__name__ in cls
        )

    @staticmethod
    def is_a_class(
        element: type
    ):
        """
        Check if the provided 'element' is a class.
        """
        return PythonValidator.is_instance(element, type)
        # This below is just another alternative I want to keep
        return getattr(element, '__name__', None) is not None
    
    @staticmethod
    def is_subclass(
        element: type,
        cls: Union[str, type]
    ) -> bool:
        """
        Check if the provided 'element' is a subclass of the provided
        class 'cls'. The 'cls' parameter can be the class or the
        string name of that class.

        You can pass the string name of the class to avoid import in 
        the file where you are calling this method.

        This method can return True if the provided 'element' is an
        instance or a class that inherits from Enum or YTAEnum class.
        Consider this above to check later if your 'element' is an
        instance or a class.
        """
        # TODO: Convert this to accept cls as array and check 
        # diferent classes
        if (
            not PythonValidator.is_string(cls) and
            not PythonValidator.is_a_class(cls)
        ):
            raise Exception(f'The provided "cls" parameter "{str(cls)}" is not a string nor a class.')
        
        if not PythonValidator.is_a_class(element):
            # We want to know if it is a subclass, we don't
            # care if the provided one is an instance or a class,
            # so we try with both.
            if PythonValidator.is_an_instance(element):
                element = type(element)
            else:
                return False
            
        """
        The element.__mro__ returns a list like this below in
        which you can see, oredered, the classes hierarchy.
        This can be useful, so I keep it here:

        (<class 'yta_multimedia.video.generation.manim.classes.text.simple_text_manim_animation.ExampleManimAnimationGenerator'>, <class 'yta_multimedia.video.generation.manim.classes.base_manim_animation.BaseManimAnimation'>, <class 'manim.scene.scene.Scene'>, <class 'object'>)    
        """
            
        return (
            cls in [
                base_class.__name__
                for base_class in element.__bases__
            ]
            if PythonValidator.is_string(cls) else
            issubclass(element, cls)
        )

    @staticmethod
    def is_a_function(
        element: 'function'
    ) -> bool:
        """
        Check if the provided 'element' is a function.
        """
        # TODO: Maybe inspect.isfunction(element) (?)
        return type(element).__name__ == 'function'
    
    @staticmethod
    def is_class_staticmethod(
        cls: type,
        method: 'function',
        method_name: str = None
    ) -> bool:
        """
        Check if the provided 'method' is an staticmethod (a
        function) defined in the also provided 'cls' class.

        If the 'method_name' parameter is provided, it will
        also check if the name of the provided 'method' is
        equal to the one provided in the 'method_name' param.
        """
        if not PythonValidator.is_a_class(cls):
            raise Exception('The provided "cls" parameter is not a class.')
        
        if not PythonValidator.is_a_function(method):
            raise Exception('The provided "method" parameter is not a function.')
        
        for function in inspect.getmembers(cls, predicate = inspect.isfunction):
            if function[0] == method.__name__:
                return (
                    True                            # It is one staticmethod
                    if method_name is None else
                    method.__name__ == method_name  # It is the specific staticmethod
                )
            
        return False

    @staticmethod
    def do_dict_has_keys(
        element: dict,
        keys: list[str]
    ) -> bool:
        """
        Check if the provided 'element' dict has the also
        given 'keys'.
        """
        return all(key in keys for key in element)

    @staticmethod
    def get_missing_keys_in_dict(
        element: dict,
        keys: list[str]
    ) -> list[str]:
        """
        Get the keys in the provided 'keys' that are missing
        in the also given 'element' dict.
        """
        return list(set(keys) - set(element.keys()))
    
    @staticmethod
    def is_list(
        element: list
    ) -> bool:
        """
        Check if the provided 'element' is a list, which
        is quite different from an array.
        """
        return type(element) == list
    
    @staticmethod
    def is_array(
        element: array
    ) -> bool:
        """
        Check if the provided 'element' is an array, which
        is quite different from a list.
        """
        return type(element) == array

    @staticmethod
    def is_empty_list(
        element: list
    ) -> bool:
        """
        Check if the provided 'element' is a list but empty.
        """        
        return (
            PythonValidator.is_list(element) and
            len(element) == 0
        )
    
    @staticmethod
    def is_list_of_instances(
        element: list[object]
    ) -> bool:
        """
        Check if the provided 'element' is a list in which
        all the items are instances of any class.
        """
        return (
            PythonValidator.is_list(element) and
            all(
                PythonValidator.is_an_instance(item)
                for item in element
            )
        )
    
    @staticmethod
    def is_list_of_these_instances(
        element: list[object],
        cls: type
    ) -> bool:
        """
        Check if the provided 'element' is a list in which
        all the items are instances of the provided 'cls'
        class.
        """
        return (
            PythonValidator.is_list(element) and
            all(
                PythonValidator.is_instance(item, cls)
                for item in element
            )
        )
    
    @staticmethod
    def is_list_of_string(
        element: list[str]
    ) -> bool:
        """
        Check if the provided 'element' is a list in which
        all the items are strings.
        """
        return (
            PythonValidator.is_list(element) and
            all(
                PythonValidator.is_string(item)
                for item in element
            )
        )

    def is_list_of_numbers(
        element: list[Union[int, float]]
    ) -> bool:
        """
        Check if the provided 'element' is a list in which
        all the items are numbers.
        """
        return (
            PythonValidator.is_list(element) and
            all(
                PythonValidator.is_number(item)
                for item in element
            )
        )

    def is_list_of_positive_numbers(
        element: list[Union[int, float]],
        do_include_zero: bool = True
    ) -> bool:
        """
        Check if the provided 'element' is a list in which
        all the items are positive numbers.
        """
        from yta_general_utils.programming.validator.number import NumberValidator

        return (
            PythonValidator.is_list(element) and
            all(
                NumberValidator.is_positive_number(item, do_include_zero = do_include_zero)
                for item in element
            )
        )
    
    @staticmethod
    def is_list_of_classes(
        element: list[type]
    ) -> bool:
        """
        Check if the provided 'element' is a list in which
        all the items are classes.
        """
        return (
            PythonValidator.is_list(element) and
            all(
                PythonValidator.is_a_class(item)
                for item in element
            )
        )

    @staticmethod
    def is_list_of_these_classes(
        element: list[type],
        cls: Union[list[Union[type, str]], str, type]
    ) -> bool:
        """
        Check if the provided 'element' is a list in which
        all the items are classes from the given 'cls' 
        class or class array.
        """
        return (
            PythonValidator.is_list(element) and
            all(
                PythonValidator.is_class(item, cls)
                for item in element
            )
        )

    @staticmethod
    def is_dict(
        element: dict
    ) -> bool:
        """
        Check if the provided 'element' is a dict.
        """
        return PythonValidator.is_instance(element, dict)

    @staticmethod
    def is_tuple(
        element
    ) -> bool:
        """
        Check if the provided 'element' is a tuple.
        """
        return PythonValidator.is_instance(element, tuple)
    
    @staticmethod
    def is_tuple_or_list_or_array_of_n_elements(
        element: Union[tuple, list, array],
        n: int
    ) -> bool:
        """
        Check if the provided 'element' is a tuple or a list
        with 'n' values.
        """
        return (
            (
                PythonValidator.is_tuple(element) or 
                PythonValidator.is_list(element) or
                PythonValidator.is_array(element)
            ) and
            len(element) == n
        )
    
    @staticmethod
    def is_string(
        element: str
    ) -> bool:
        """
        Check if the provided 'element' is a string (str).
        """
        return PythonValidator.is_instance(element, str)
    
    @staticmethod
    def is_boolean(
        element: bool
    ) -> bool:
        """
        Check if the provided 'element' is a boolean (bool).
        """
        return PythonValidator.is_instance(element, bool)
    
    @staticmethod
    def is_number(
        element: Union[int, float, str],
        do_accept_string_number: bool = False
    ) -> bool:
        """
        Check if the provided 'element' is a numeric value. If
        'do_accept_string_number' is True, it will try to parse
        the 'element' as a float if a string is provided.
        """
        from yta_general_utils.programming.validator.number import NumberValidator

        return NumberValidator.is_number(element, do_accept_string_number)
    
    @staticmethod
    def is_numpy_array(
        element: np.ndarray
    ) -> bool:
        """
        Check if the provided 'element' is an instance of the
        numpy array np.ndarray.
        """
        return PythonValidator.is_instance(element, np.ndarray)

    @staticmethod
    def is_url(
        element: str
    ) -> bool:
        """
        Check if the provided 'element' is a valid url.
        """
        return validators.url(element)
    
    @staticmethod
    def is_enum(
        element: Union['YTAEnum', Enum]
    ) -> bool:
        """
        Check if the provided 'element' is a subclass of an Enum or
        a YTAEnum.

        This method can return True if the provided 'element' is an
        instance or a class that inherits from Enum or YTAEnum class.
        Consider this above to check later if your 'element' is an
        instance or a class.
        """
        # TODO: I think it is 'EnumMeta' not Enum
        return (
            PythonValidator.is_subclass(element, 'YTAEnum') or
            PythonValidator.is_subclass(element, 'Enum')
        )
    
    @staticmethod
    def is_enum_instance(
        element: Union['YTAEnum', Enum]
    ) -> bool:
        """
        Check if the provided 'element' is a Enum (it is a subclass
        of an Enum or a YTAEnum) and it is an instance.
        """
        return (
            PythonValidator.is_enum(element) and
            PythonValidator.is_an_instance(element)
        )
    
    @staticmethod
    def is_enum_class(
        element: Union['YTAEnum', Enum]
    ) -> bool:
        """
        Check if the provided 'element' is a Enum (it is a subclass
        of an Enum or a YTAEnum) and it is a class.
        """
        return (
            PythonValidator.is_enum(element) and
            PythonValidator.is_a_class(element)
        )
    
    @staticmethod
    def is_callable(
        element: callable
    ) -> bool:
        """
        Check if the provided 'element' is a callable
        instance (method, function) or not.
        """
        return callable(element)
    
    # TODO: This method below is not complete, need work
    @staticmethod
    def validate_method_params(
        method: 'function',
        params: list, 
        params_to_ignore: list[str] = ['self', 'cls', 'args', 'kwargs']
    ):
        """
        IMPORTANT! This method should be called on top of any method
        in which you want to validate if the provided parameters are
        valid, by providing the method (function) declaration and
        also the 'locals()' function executed as 'params' parameter.
        So, it should be called like this:

        PythonValidator.validate_method_params(function, locals())

        This method check the types of the params that the provided
        'method' has and validate if the provided values fit the
        specified types (according also to the default values). It
        will raise an Exception when any of the provided params (and
        not ignored) is not valid according to its type declaration.

        The provided 'params' must be a dict containing all the param
        names and values.

        This method is able to parse non-type declarations, as in
        "method(type)", single declarations, as in "method(type: str)"
        and Union declarations, as in
        "method(type: Union[str, MyClass])"

        The 'method' parameter must be a real python method to be able
        to inspect it.
        """
        if not PythonValidator.is_a_function(method):
            raise Exception('The provided "method" parameter is not a valid method (function).')

        # This below is like '<class 'package.of.class.name'> or '<class 'str'>
        SINGLE_TYPE_REGEX = r"<class '([^']+)'>"
        # This below is like 'Union['str', 'int', 'FfmpegHandler']
        UNION_TYPE_REGEX = r"typing\.Union\[\s*((?:[^,]+(?:\s*,\s*)?)+)\s*\]"
        # This below is to flag those params with no default value
        # because None can be a default value indicating that it is
        # an optional value
        NO_DEFAULT_VALUE = '__no_default_value__'
        NO_TYPE = '__no_type__'

        # TODO: Refactor this below to make it easier to be read
        for param in inspect.signature(method).parameters.values():
            if param.name in params_to_ignore:
                continue

            print(param.name)
            print(params_to_ignore)

            types = param.annotation if param.annotation is not inspect.Parameter.empty else NO_TYPE
            default_value = param.default if param.default is not inspect.Parameter.empty else NO_DEFAULT_VALUE

            # 'types' can be nothing, a single type or an Union
            if types:
                match_class = re.match(SINGLE_TYPE_REGEX, str(types))
                match_union = re.match(UNION_TYPE_REGEX, str(types))

                # Turn type to array of string types
                if match_class:
                    types = [match_class.group(1).split('.')[-1]]
                elif match_union:
                    classes = match_union.group(1).split(',')
                    types = [class_i.strip().split('.')[-1] for class_i in classes]

            # Now check with the param provided
            user_param = params.get(param.name, None)

            if types is NO_TYPE and user_param is None:
                # If no type we cannot validate anything, but I 
                # think if no type nor value it will not be
                # executed and this Exception below will never
                # happen
                raise Exception(f'The param "{str(param.name)}" has no type declaration but also None value, so it is not accepted.')
            elif user_param == None and (PythonValidator.is_list(types) and 'None' in types or default_value == None):
                # TODO: If we are strict with typing, a value that
                # can be None should be Union[None, ...] and also
                # param = None (as default value) to indicate it is
                # optional, but we accept both of them separately

                # Param value is None and None is accepted or is
                # default value
                pass
            elif user_param == None:
                raise Exception(f'The param "{str(param.name)}" has None value provided and we expected one of these types: {", ".join(types)}.')
            else:
                if not PythonValidator.is_instance(user_param, types) and types is not NO_TYPE:
                    print(types)
                    types_str = ', '.join(types)
                    raise Exception(f'The param value "{str(param.name)}" provided "{str(user_param)}" is not one of the expected types: {types_str}.')

        return True



from yta_general_utils.file.remover import FileRemover
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.programming.env import Environment
from random import randint
from datetime import datetime

import os


class Temp:
    """
    Class to encapsulate the temp files functionality.
    """

    # TODO: I don't like the way we handle this variable
    # so please, refactor and improve it :). I would like
    # it to be a class/static property to use .WIP_FOLDER
    @staticmethod
    def WIP_FOLDER():
        """
        The Work In Progress folder in which we store all
        the temporary files we are working with. Its value
        is loaded from the '.env' 'WIP_FOLDER' variable.
        """
        # TODO: Cyclic import issue
        from yta_general_utils.file.checker import FileValidator

        WIP_FOLDER = Environment.get_current_project_env('WIP_FOLDER')
        if not WIP_FOLDER:
            # We force creating the dir
            WIP_FOLDER = get_project_abspath() + 'yta_wip/'
            if not FileValidator.is_folder(WIP_FOLDER):
                os.mkdir(WIP_FOLDER)

        return WIP_FOLDER

    # @staticmethod
    # def _initialize() -> str:
    #     """
    #     Load the current project dotenv and look for the 
    #     'WIP_FOLDER' variable definition in that file and,
    #     if not defined, set it with the default 'yta_wip'
    #     value and creates the folder if it doesn't exist.

    #     This method exist to make sure the temporary folder
    #     exist before working with it.
    #     """
    #     WIP_FOLDER = Environment.get_current_project_env('WIP_FOLDER')
    #     if not WIP_FOLDER:
    #         # We force creating the dir
    #         WIP_FOLDER = get_project_abspath() + 'yta_wip/'
    #         if not FileValidator.is_folder(WIP_FOLDER):
    #             os.mkdir(WIP_FOLDER)

    #     return WIP_FOLDER
    
    # TODO: Rename these 2 methods below please, they
    # are not very self-descriptive
    @staticmethod
    def get_filename(
        filename: str
    ) -> str:
        """
        Get a temporary file name using the given 'filename'
        and including a random suffix related to the current
        datetime. This is just a filename that doesn't 
        include the temporary folder or any prefix.

        This method uses the current datetime and a random 
        integer to be always unique.

        If you provide 'file.wav' it will return something
        similtar to 'file_202406212425.wav'.
        """
        delta = (datetime.now() - datetime(1970, 1, 1))
        aux = filename.split('.')

        return f'{aux[0]}_{str(int(delta.total_seconds()))}{str(randint(0, 10_000))}.{aux[1]}'
    
    @staticmethod
    def create_filename(
        filename: str
    ) -> str:
        """
        Get a temporary file name using the given 'filename'
        and including a random suffix related to the current
        datetime. This is a filename that includes the 
        temporary folder as a prefix so it can be used in
        the app.

        This method uses the current datetime and a random 
        integer to be always unique.

        If you provide 'file.wav' it will return something
        similtar to '$WIP_FOLDER/file_202406212425.wav'.
        """
        # TODO: Rename this as it uses wip and we do not mention it
        # TODO: Issue if no extension provided
        return Temp.create_custom_filename(Temp.get_filename(filename))
    
    @staticmethod
    def create_custom_filename(
        filename: str
    ) -> str:
        """
        Get a file name that includes the 'WIP_FOLDER' as
        a prefix but preserves the provided 'filename'. 
        This is useful when we need a temporary file to
        work with but with a specific name, maybe not for
        a long time but able to access it from different
        points of the app because of its known (and not
        random) name.
        """
        return f'{Temp.WIP_FOLDER()}{filename}'

    @staticmethod
    def clean_folder() -> None:
        """
        Remove all the existing files in the temporary 
        folder.
        """
        FileRemover.delete_files(Temp.WIP_FOLDER())


from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from yta_general_utils.web.scraper.chrome_scraper import go_to_and_wait_loaded, start_chrome

import time
import os
import pyautogui
import glob
import shutil

MESSAGE_SENDER_ME = 1
MESSAGE_SENDER_OTHER = 2
PROJECT_ABSOLUTE_PATH = os.getenv('PROJECT_ABSOLUTE_PATH')
TOUSE_ABSOLUTE_PATH = os.getenv('TOUSE_ABSOLUTE_PATH')
# TODO: Move this to .env variable
DOWNLOADS_ABSOLUTE_PATH = 'C:/Users/dania/Downloads/'

"""
inputs = driver.find_elements(By.TAG_NAME, 'input')
profile1_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person1")]')
profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person2")]')
# One checkbox below
is_verified = driver.find_element(By.ID, 'tiktokverfied')
time_input = driver.find_element(By.ID, 'edit-2-Clock')
operator_input = driver.find_element(By.ID, 'iphoneoperator')
battery_input = driver.find_element(By.ID, 'Battery_percent')
profile1_name_input = driver.find_element(By.ID, 'profile1_name')
date_input = driver.find_element(By.ID, 'time-profile1')
# Two checkboxes below
is_story_active_checkbox = driver.find_element(By.ID, 'visibleStory')
is_online_checkbox = driver.find_element(By.ID, 'visibleOnline')
is_online_checkbox.click()
time.sleep(1)
# Two radio buttons below plus one text when 'online_time_is_visible'
online_time_or_nickname_is_hidden = driver.find_element(By.ID, 'Active_hide')
online_time_or_nickname_is_visible = driver.find_element(By.CSS_SELECTOR, 'input#Active_show')
online_time_or_nickname_text = driver.find_element(By.ID, 'edit-active')
"""

def __setup(driver, name, username, profile_picture_absolute_path):
    # TODO: Set maximum 'name' and 'username' lengths
    __set_name(driver, name)
    if username:
        __set_username(driver, username)
    __set_profile_picture(driver, profile_picture_absolute_path)

def __set_profile_picture(driver, absolute_path):
    element_present = EC.presence_of_element_located((By.ID, 'profileImg'))
    element = WebDriverWait(driver, 10).until(element_present)
    time.sleep(1)
    element = element.find_element(By.XPATH, '..')
    element.click()
    time.sleep(2)
    pyautogui.write((absolute_path).replace('/', '\\')) 
    time.sleep(1)
    pyautogui.press('enter')
    time.sleep(1)

def __activate_blue_check(driver):
    is_verified_checkbox = driver.find_element(By.ID, 'tiktokverfied')
    if not is_verified_checkbox.is_selected():
        is_verified_checkbox.click()

def __deactivate_blue_check(driver):
    is_verified_checkbox = driver.find_element(By.ID, 'tiktokverfied')
    if is_verified_checkbox.is_selected():
        is_verified_checkbox.click()

def __activate_story_active(driver):
    is_story_active_checkbox = driver.find_element(By.ID, 'visibleStory')
    if not is_story_active_checkbox.is_selected():
        is_story_active_checkbox.click()

def __deactivate_story_active(driver):
    is_story_active_checkbox = driver.find_element(By.ID, 'visibleStory')
    if is_story_active_checkbox.is_selected():
        is_story_active_checkbox.click()

def __activate_is_online(driver):
    is_online_checkbox = driver.find_element(By.ID, 'visibleOnline')
    if not is_online_checkbox.is_selected():
        is_online_checkbox.click()

def __deactivate_is_online(driver):
    is_online_checkbox = driver.find_element(By.ID, 'visibleOnline')
    if is_online_checkbox.is_selected():
        is_online_checkbox.click()

def __set_conversation_date(driver, str):
    date_input = driver.find_element(By.ID, 'time-profile1')
    time.sleep(1)
    date_input.clear()
    date_input.send_keys(str)

def __set_name(driver, name):
    profile1_name_input = driver.find_element(By.ID, 'profile1_name')
    time.sleep(1)
    profile1_name_input.clear()
    profile1_name_input.send_keys(name)

def __set_username(driver, username):
    __deactivate_is_online(driver)
    time.sleep(1)
    show_username_radiobutton = driver.find_element(By.CSS_SELECTOR, 'input#Active_show')
    if not show_username_radiobutton.is_selected():
        show_username_radiobutton.click()
    time.sleep(1)
    # edit-active
    time_or_nickname_input = driver.find_element(By.ID, 'edit-active')
    time.sleep(1)
    time_or_nickname_input.clear()
    time_or_nickname_input.send_keys(username)

def __download(driver, output_filename):
    time.sleep(2)
    download_button = driver.find_element(By.ID, 'snapshot')
    download_button.click()
    time.sleep(10)
    # Move downloaded file to our destination
    list_of_files = glob.glob(DOWNLOADS_ABSOLUTE_PATH + '*.png')
    latest_file = max(list_of_files, key = os.path.getctime)
    shutil.move(latest_file, os.path.join(PROJECT_ABSOLUTE_PATH + output_filename))

def __write_my_message(driver, message):
    driver.execute_script('window.scrollTo(0, 0)')
    profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person2")]')
    profile2_tab.click()
    time.sleep(1)
    profile2_textarea = driver.find_element(By.ID, 'profile2_message')
    profile2_textarea.clear()
    profile2_textarea.send_keys(message)
    time.sleep(1)
    actions = ActionChains(driver)
    for i in range(2):
        actions.send_keys(Keys.TAB)
    actions.send_keys(Keys.SPACE)
    actions.perform()
    time.sleep(1)

def __write_other_message(driver, message):
    driver.execute_script('window.scrollTo(0, 0)')
    profile1_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person1")]')
    profile1_tab.click()
    time.sleep(1)
    profile1_textarea = driver.find_element(By.ID, 'profile1_message')
    profile1_textarea.clear()
    profile1_textarea.send_keys(message)
    time.sleep(1)
    actions = ActionChains(driver)
    actions.send_keys(Keys.TAB)
    actions.send_keys(Keys.SPACE)
    actions.perform()
    time.sleep(1)

# TODO: Implement sending foto

def generate_conversation(name, username, messages, profile_picture_absolute_path, date_str, output_filename):
    # Messages must come with autor and text
    URL = 'https://prankshit.com/fake-instagram-chat-generator.php'

    profile_picture_absolute_path = profile_picture_absolute_path.replace('/', '\\')

    try:
        driver = start_chrome(True)
        go_to_and_wait_loaded(driver, URL)

        time.sleep(2)
        __setup(driver, name, username, profile_picture_absolute_path)
        time.sleep(1)

        # Add more data ('active')
        messages = [
            {
                'sender': MESSAGE_SENDER_ME,
                'message': 'Holaaaaa!'
            },
            {
                'sender': MESSAGE_SENDER_ME,
                'message': 'estás??'
            },
            {
                'sender': MESSAGE_SENDER_OTHER,
                'message': 'Sí, ¿qué pasa wacho??'
            },
            {
                'sender': MESSAGE_SENDER_ME,
                'message': 'Nada, solo quería saludarte!! Campeón!'
            },
            {
                'sender': MESSAGE_SENDER_OTHER,
                'message': 'Dalee! Grande, gracias!!'
            },
        ]

        for index, message in enumerate(messages):
            if message['sender'] == MESSAGE_SENDER_ME:
                __write_my_message(driver, message['message'])
            else:
                __write_other_message(driver, message['message'])

            __download(driver, 'wip/test_instagram_' + str(index) + '.png')
    finally:
        driver.close()


"""
Thank you:

https://github.com/Kemaweyan/singleton_decorator
"""
import functools


class _SingletonWrapper:
    """
    Singleton wrapper class that creates instances
    for each decorated class.
    """

    def __init__(
        self,
        cls
    ):
        self.__wrapped__ = cls
        self._instance = None

        functools.update_wrapper(self, cls)

    def __call__(
        self,
        *args,
        **kwargs
    ):
        """
        Get a single instance of decorated class.
        """
        if self._instance is None:
            self._instance = self.__wrapped__(*args, **kwargs)

        return self._instance



from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.text.transformer import remove_marks_and_accents
from dataclasses import dataclass


@dataclass
class TermFound:
    """
    @dataclass
    A term found in a text, containing the indexes
    in which the term was found.
    """

    term: str
    """
    The term that has been found.
    """
    start_index: int
    """
    The index (within the text) in which the start
    of the 'term' has been found.
    """
    end_index: int
    """
    The index (within the text) in which the end
    of the 'term' has been found.
    """

    def __init__(
        self,
        term: str,
        start_index: int,
        end_index: int
    ) -> 'TermFound':
        ParameterValidator.validate_mandatory_string('term', term, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_int('start_index', start_index, do_include_zero = True)
        ParameterValidator.validate_mandatory_positive_int('end_index', end_index, do_include_zero = True)

        self.term = term
        self.start_index = start_index
        self.end_index = end_index


class TextFinderMode(Enum):
    """
    This is the mode in which we will look for the terms
    in the given segment text to find any coincidences.
    """

    EXACT = 'exact'
    """
    The term found must be exactly matched on the text,
    which means that accents and punctuation marks will
    be considered.
    """
    IGNORE_CASE_AND_ACCENTS = 'ignore_case_and_accents'
    """
    The term found must match, in lower case and ignoring
    the accents, the text.
    """

class TextFinder:
    """
    Class to simplify the way we look for texts
    within other texts.
    """

    @staticmethod
    def find_in_text(
        term: str,
        text: str,
        mode: TextFinderMode = TextFinderMode.IGNORE_CASE_AND_ACCENTS
    ) -> list[TermFound]:
        """
        Find the provided 'term' in the also provided
        'text' and obtain the start and end indexes
        of the 'term' according to that 'text'. The 
        term can be more than once in the 'text'.

        TODO: Add an ilustrative example.

        This method returns an array containing
        TermFound instances including the start and
        end indexes.
        """
        ParameterValidator.validate_mandatory_string('term', term, do_accept_empty = False)
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = True)
        
        mode = TextFinderMode.to_enum(mode)

        if mode == TextFinderMode.IGNORE_CASE_AND_ACCENTS:
            term = remove_marks_and_accents(term).lower()
            text = remove_marks_and_accents(text).lower()

        text_words = text.split()
        term_words = term.split()
        
        # Store first and last index if found
        return [
            TermFound(
                term = term,
                start_index = i,
                end_index = i + len(term_words) - 1
            )
            for i in range(len(text_words) - len(term_words) + 1)
            if text_words[i:i + len(term_words)] == term_words
        ]


"""
This is a general constants file in which we
want to keep different constants in a general
context so we can use these ones to pass them
to any other library we create in order to
obtain it corresponding value.

For example, we have a Language constant in
which we store the different languages 
available in the world. But, of course, Google
will treat those languages in one way and 
other libraries will do it different. We, when
creating our own libraries, will create a 
method that allows us passing this general
constant to obtain its corresponding of that
specific library. Why? Because if we have a
frontend that interacts with 3 different 
libraries, we cannot let the user choose 
between 3 different type of language constants,
we need only one and then it is turned into
its specific value.
"""
from yta_general_utils.programming.enum import YTAEnum as Enum


class Language(Enum):
    """
    All the languages of the world in a general
    context. Use this enum class to convert it
    to your library-specific language.

    This is based on the ISO-639 language code
    standard: https://pypi.org/project/iso639-lang/

    Maybe I should import the library and let it
    be used like 'Lang("es")' to autodetect it and
    transform to our Enum values (?).
    """

    ABKHAZIAN = 'ab'
    AFAR = 'aa'
    AFRIKAANS = 'af'
    AKAN = 'ak'
    ALBANIAN = 'sq'
    AMHARIC = 'am'
    ARABIC = 'ar'
    ARAGONESE = 'an'
    ARMENIAN = 'hy'
    ASSAMESE = 'as'
    AVARIC = 'av'
    AVESTAN = 'ae'
    AYMARA = 'ay'
    AZERBAIJANI = 'az'
    BAMBARA = 'bm'
    BASHKIR = 'ba'
    BASQUE = 'eu'
    BELARUSIAN = 'be'
    BENGALI = 'bn'
    BISLAMA = 'bi'
    BOSNIAN = 'bs'
    BRETON = 'br'
    BULGARIAN = 'bg'
    BURMESE = 'my'
    CATALAN = 'ca'
    CHAMORRO = 'ch'
    CHECHEN = 'ce'
    CHICHEWA = 'ny'
    CHINESE = 'zh'
    CHINESE_TRADITIONAL = 'zh-Hans'
    # TODO: I think there are more languages like
    # this missing
    CHURCH_SLAVONIC = 'cu'
    CHUVASH = 'cv'
    CORNISH = 'kw'
    CORSICAN = 'co'
    CREE = 'cr'
    CROATIAN = 'hr'
    CZECH = 'cs'
    DANISH = 'da'
    DIVEHI = 'dv'
    DUTCH = 'nl'
    DZONGKHA = 'dz'
    ENGLISH = 'en'
    ESPERANTO = 'eo'
    ESTONIAN = 'et'
    EWE = 'ee'
    FAROESE = 'fo'
    FIJIAN = 'fj'
    FINNISH = 'fi'
    FRENCH = 'fr'
    WESTERN_FRISIAN = 'fy'
    FULAH = 'ff'
    GAELIC = 'gd'
    GALICIAN = 'gl'
    GANDA = 'lg'
    GEORGIAN = 'ka'
    GERMAN = 'de'
    GREEK = 'el'
    KALAALLISUT = 'kl'
    GUARANI = 'gn'
    GUJARATI = 'gu'
    HAITIAN = 'ht'
    HAUSA = 'ha'
    HEBREW = 'he'
    HERERO = 'hz'
    HINDI = 'hi'
    HIRI_MOTU = 'ho'
    HUNGARIAN = 'hu'
    ICELANDIC = 'is'
    IDO = 'io'
    IGBO = 'ig'
    INDONESIAN = 'id'
    INTERLINGUA = 'ia'
    INTERLINGUE = 'ie'
    INUKTITUT = 'iu'
    INUPIAQ = 'ik'
    IRISH = 'ga'
    ITALIAN = 'it'
    JAPANESE = 'ja'
    JAVANESE = 'jv'
    KANNADA = 'kn'
    KANURI = 'kr'
    KASHMIRI = 'ks'
    KAZAKH = 'kk'
    CENTRAL_KHMER = 'km'
    KIKUYU = 'ki'
    KINYARWANDA = 'rw'
    KYRGYZ = 'ky'
    KOMI = 'kv'
    KONGO = 'kg'
    KOREAN = 'ko'
    KUANYAMA = 'kj'
    KURDISH = 'ku'
    LAO = 'lo'
    LATIN = 'la'
    LATVIAN = 'lv'
    LIMBURGAN = 'li'
    LINGALA = 'ln'
    LITHUANIAN = 'lt'
    LUBA_KATANGA = 'lu'
    LUXEMBOURGISH = 'lb'
    MACEDONIAN = 'mk'
    MALAGASY = 'mg'
    MALAY = 'ms'
    MALAYALAM = 'ml'
    MALTESE = 'mt'
    MANX = 'gv'
    MAORI = 'mi'
    MARATHI = 'mr'
    MARSHALLESE = 'mh'
    MONGOLIAN = 'mn'
    NAURU = 'na'
    NAVAJO = 'nv'
    NORTH_NDEBELE = 'nd'
    SOUTH_NDEBELE = 'nr'
    NDONGA = 'ng'
    NEPALI = 'ne'
    NORWEGIAN = 'no'
    NORWEGIAN_BOKMAL = 'nb'
    NORWEGIAN_NYNORSK = 'nn'
    OCCITAN = 'oc'
    OJIBWA = 'oj'
    ORIYA = 'or'
    OROMO = 'om'
    OSSETIAN = 'os'
    PALI = 'pi'
    PASHTO = 'ps'
    PERSIAN = 'fa'
    POLISH = 'pl'
    PORTUGUESE = 'pt'
    PUNJABI = 'pa'
    QUECHUA = 'qu'
    ROMANIAN = 'ro'
    ROMANSH = 'rm'
    RUNDI = 'rn'
    RUSSIAN = 'ru'
    NORTHERN_SAMI = 'se'
    SAMOAN = 'sm'
    SANGO = 'sg'
    SANSKRIT = 'sa'
    SARDINIAN = 'sc'
    SERBIAN = 'sr'
    SHONA = 'sn'
    SINDHI = 'sd'
    SINHALA = 'si'
    SLOVAK = 'sk'
    SLOVENIAN = 'sl'
    SOMALI = 'so'
    SOUTHERN_SOTHO = 'st'
    SPANISH = 'es'
    SUNDANESE = 'su'
    SWAHILI = 'sw'
    SWATI = 'ss'
    SWEDISH = 'sv'
    TAGALOG = 'tl'
    TAHITIAN = 'ty'
    TAJIK = 'tg'
    TAMIL = 'ta'
    TATAR = 'tt'
    TELUGU = 'te'
    THAI = 'th'
    TIBETAN = 'bo'
    TIGRINYA = 'ti'
    TONGA = 'to'
    TSONGA = 'ts'
    TSWANA = 'tn'
    TURKISH = 'tr'
    TURKMEN = 'tk'
    TWI = 'tw'
    UIGHUR = 'ug'
    UKRAINIAN = 'uk'
    URDU = 'ur'
    UZBEK = 'uz'
    VENDA = 've'
    VIETNAMESE = 'vi'
    VOLAPUK = 'vo'
    WALLOON = 'wa'
    WELSH = 'cy'
    WOLOF = 'wo'
    XHOSA = 'xh'
    SICHUAN_YI = 'ii'
    YIDDISH = 'yi'
    YORUBA = 'yo'
    ZHUANG = 'za'
    ZULU = 'zu'



"""
Module to handle the html file that allows us using
the web navigator speech recognition system.
"""
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.downloader import Downloader


TRANSCRIBER_HTML_FILENAME = 'index.html'
TRANSCRIBER_HTML_ABSPATH = f'{get_project_abspath()}{TRANSCRIBER_HTML_FILENAME}'

def download_web_file():
    """
    Download the html file from Google Drive if
    not available locally.
    """
    if not FileValidator.file_exists(TRANSCRIBER_HTML_ABSPATH):
        Downloader.download_google_drive_resource(
            'https://drive.google.com/file/d/1KQs6D7Zmd2Oj7mT4JTV8S38e2ITu_gUs/view?usp=sharing',
            TRANSCRIBER_HTML_FILENAME
        )


from yta_general_utils.web.scraper.chrome_scraper import ChromeScraper
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.text.finder import TextFinder
from yta_general_utils.text.transcriptor.web import TRANSCRIBER_HTML_ABSPATH, download_web_file
from yta_general_utils.text.transformer import remove_non_ascii_characters
from typing import Union

import time


class WebRealTimeAudioTranscriptor:
    """
    Class to wrap a functionality related to real
    time audio transcription by using a web scraper.

    This class uses local files to create a simple
    web page that uses the chrome speech recognition
    to get the transcription.
    """

    _LOCAL_URL: str = f'file:///{TRANSCRIBER_HTML_ABSPATH}'
    _REMOTE_URL: str = 'https://iridescent-pie-f24ff0.netlify.app/'
    _WAITING_TIME: float = 0.1
    """
    The url to our local web page file.
    """
    max_time_to_wait: Union[float, None]
    """
    The maximum time the software will be waiting
    to detect an audio transcription before exiting
    with an empty result.
    """
    time_to_stop: float
    """
    The time that has to be spent once a final
    transcription has been found to consider it
    as a definitive one. There can be more final
    transcriptions after that one due to some 
    logic that I still don't understand properly.
    """
    do_use_local_web_page: bool
    """
    Flag that indicates if the resource must be a
    local web page (that will be loaded from a file
    in our system) or from a remote url.
    """

    def __init__(
        self,
        max_time_to_wait: Union[float, None] = 15.0,
        time_to_stop: float = 1.5,
        do_use_local_web_page: bool = True
    ):
        ParameterValidator.validate_positive_float('max_time_to_wait', max_time_to_wait, do_include_zero = True)
        ParameterValidator.validate_mandatory_float('time_to_stop', time_to_stop)
        ParameterValidator.validate_positive_float('time_to_stop', time_to_stop, do_include_zero = False)

        self.scraper = ChromeScraper()
        self.max_time_to_wait = (
            9999 # TODO: This is risky if no microphone or something
            if (
                max_time_to_wait == 0 or
                max_time_to_wait is None
            ) else
            max_time_to_wait
        )
        self.time_to_stop = time_to_stop
        self.do_use_local_web_page = do_use_local_web_page

        if self.do_use_local_web_page:
            # We need to make sure the file exist
            download_web_file()

        self._load()

    @property
    def url(
        self
    ) -> str:
        """
        The url that must be used to interact with the
        web page that is able to catch the audio and
        transcribe it.
        """
        return (
            self._LOCAL_URL
            if self.do_use_local_web_page else
            self._REMOTE_URL
        )
    
    @property
    def is_listening(
        self
    ) -> bool:
        """
        Indicate if the microphone is activated and the
        web page is listening to the user's speech.
        """
        self._toggle_narration_button.text == 'Click to stop'

    @property
    def _toggle_narration_button(
        self
    ) -> 'WebElement':
        """
        Get the main narration button WebElement.
        """
        return self.scraper.find_element_by_id('toggle')
    
    @property
    def _reset_narration_button(
        self
    ) -> 'WebElement':
        """
        Get the reset narration button WebElement.
        """
        return self.scraper.find_element_by_id('reset_button')

    @property
    def _final_transcription_paragraph(
        self
    ) -> 'WebElement':
        """
        Get the final transcription paragraph WebElement.
        """
        return self.scraper.find_element_by_id('final_transcription')

    @property
    def _temp_transcription_paragraph(
        self
    ) -> 'WebElement':
        """
        Get the temporary transcription paragraph WebElement.
        """
        return self.scraper.find_element_by_id('temp_transcription')

    def _load(
        self
    ):
        """
        Navigates to the web page when not yet on it.

        For internal use only.
        """
        if self.scraper.current_url != self.url:
            self.scraper.go_to_web_and_wait_until_loaded(self.url)

    def reload(
        self
    ):
        """
        Force a refresh in the web page to reload it.
        """
        #self.scraper.reload()
        # Force to wait until completely loaded
        self.scraper.go_to_web_and_wait_until_loaded(self.url)

    def get_transcription(
        self
    ) -> str:
        """
        Get the text that has been transcripted from the
        audio.
        """
        self._load()

        return self._final_transcription_paragraph.text
    
    def empty_transcription(
        self
    ) -> None:
        """
        Set the definitive transcription as an empty
        string.
        """
        self._load()

        self.scraper.set_element_inner_text(
            self._final_transcription_paragraph,
            ''
        )
    
    def get_temp_transcription(
        self
    ) -> str:
        """
        Get the text that has been temporary transcripted 
        from the audio but is not still definitive.
        """
        self._load()

        return self._temp_transcription_paragraph.text

    def _get_number_of_results(
        self
    ) -> int:
        """
        Get the amount of results that have been detected
        until the moment in which it is requested. This
        count is needed to check if the user is still
        talking or not.
        """
        self._load()

        return int(self.scraper.find_element_by_id('number_of_results').text)
    
    def activate_transcription(
        self
    ) -> None:
        """
        Click the toggle button to activate the narration
        if it is not listening.
        """
        self._load()
        
        if not self.is_listening:
            self._toggle_narration_button.click()

    def deactivate_transcription(
        self
    ) -> None:
        """
        Click the toggle button to deactivate the
        narration if it is listening.
        """
        self._load()
        
        if self.is_listening:
            self._toggle_narration_button.click()

    def reset_transcription(
        self
    ) -> None:
        """
        Click the reset button that aborts the
        recognition and resets the engine instance
        so the microphone is still active but 
        ignoring the previous context and removing
        the temporary transcription text (but not
        the definitive).
        """
        self._load()

        self._reset_narration_button.click()

    def is_text_in_temporary_transcription(
        self,
        text: str
    ) -> bool:
        """
        Check if the given 'text' has been said by looking
        for in the temporary transcription.
        """
        return len(TextFinder.find_in_text(
            text,
            self.get_temp_transcription()
        )) > 0

    def is_text_in_definitive_transcription(
        self,
        text: str
    ) -> bool:
        """
        Check if the given 'text' has been said by looking
        for in the definitive transcription.
        """
        return len(TextFinder.find_in_text(
            text,
            self.get_transcription()
        )) > 0

    def is_text_in_transcription(
        self,
        text: str
    ) -> bool:
        """
        Check if the given 'text' has been said by looking
        for in the temporary and definitive transcription.
        """
        return (
            self.is_text_in_temporary_transcription(text) or
            self.is_text_in_definitive_transcription(text)
        )

    def detect_text(
        self,
        text: str
    ) -> bool:
        """
        Detect the given 'text' in the temporary or
        definitive audio transcription.
        
        This method doesn't load the web page nor click
        any button, it has to be made manually.
        """
        final_transcription = self.get_transcription()
        temp_transcription = self.get_temp_transcription()

        while (
            len(TextFinder.find_in_text(text, final_transcription)) == 0 and
            len(TextFinder.find_in_text(text, temp_transcription)) == 0
            # TODO: Append a time limit please
        ):
            time.sleep(self._WAITING_TIME)
            final_transcription = self.get_transcription()
            temp_transcription = self.get_temp_transcription()

        return True

    def detect_fast(
        self,
        text: str
    ) -> bool:
        """
        This method will load the web page and will
        activate the audio transcription and check
        the temporary results until the provided 
        'text' appears, and a True will be 
        inmediately returned in that case.

        This is called 'detect_fast' because of
        the way the web pages processed the audio
        narration. Temporary results are detected
        almost as fast as the user speaks, but the
        definitive ones last a few seconds to 
        appear. This will be checking the temporary
        results, thats why it is fast.

        TODO: By now, the 'text' is forced to be
        lowercase and without any exclamation mark,
        comma, etc.
        """
        # TODO: Refactor and improve the way we handle
        # this by checking '.', ',', etc.
        # Text must be lowercase with accents
        text = remove_non_ascii_characters(text.lower(), do_remove_accents = True)

        self.activate_transcription()

        final_transcription = self.get_transcription()
        temp_transcription = self.get_temp_transcription()

        while (
            len(TextFinder.find_in_text(text, final_transcription)) == 0 and
            len(TextFinder.find_in_text(text, temp_transcription)) == 0
            # TODO: Append a time limit please
        ):
            time.sleep(self._WAITING_TIME)
            final_transcription = self.get_transcription()
            temp_transcription = self.get_temp_transcription()

        self.deactivate_transcription()

        return True
    
    def transcribe(
        self
    ) -> str:
        """
        A web scraper instance loads the internal web
        that uses the Chrome speech recognition to get
        the audio transcription, by pressing the
        button, waiting for audio input through the
        microphone, and pressing the button again.

        If the page was previously loaded it won't be
        loaded again.

        This method will wait until no new temporary
        nor definitive results are being analyzed and
        returned to the web page, and will return the
        last definitive result found.
        """
        self.activate_transcription()

        time_elapsed = 0
        final_transcription_time_elapsed = 0
        transcription = ''
        number_of_results = 0
        while (
            time_elapsed < self.max_time_to_wait and
            (
                (
                    final_transcription_time_elapsed != 0 and
                    (final_transcription_time_elapsed + self.time_to_stop) > time_elapsed
                ) or
                final_transcription_time_elapsed == 0
            )
        ):
            tmp_final_transcription = self.get_transcription()
            tmp_number_of_results = self._get_number_of_results()
            """
            If temporary transcription is changing,
            we are still getting audio transcripted
            so we need to keep waiting for the 
            final transcription. We are indicating
            the amount of words detected, so if that
            number keeps increasing, we need to keep
            waiting
            """
            if (
                tmp_number_of_results > number_of_results or
                tmp_final_transcription != transcription
            ):
                # If final transcription has changed or the
                # amount of events keeps increasing, we 
                # keep waiting
                final_transcription_time_elapsed = time_elapsed

                if tmp_number_of_results > number_of_results:
                    number_of_results = tmp_number_of_results

                if tmp_final_transcription != transcription:
                    transcription = tmp_final_transcription
                
            time.sleep(self._WAITING_TIME)
            time_elapsed += self._WAITING_TIME

        self.deactivate_transcription()

        return transcription



from yta_general_utils.coordinate.coordinate import Coordinate
from yta_general_utils.math.rate_functions.rate_function_argument import RateFunctionArgument
from yta_general_utils.math.rate_functions.rate_function import RateFunction
from yta_general_utils.math.normalizable_value import NormalizableValue
from yta_general_utils.programming.validator.number import NumberValidator
from yta_general_utils.math.progression import Progression


# This class is very similar to PairOfNodes, maybe
# we can do something to avoid the duplicated code
class PairOfCoordinates:
    """
    Class to represent a par of coordinates that
    belong to a Map instance.
    """

    first_coordinate: Coordinate = None
    second_coordinate: Coordinate = None
    rate_function: RateFunctionArgument = None

    @property
    def max_x(self):
        """
        The maximum X value, which can be the left coordinate
        or right coordinate X value.
        """
        return max([self.first_coordinate.x, self.second_coordinate.x])
    
    @property
    def min_x(self):
        """
        The minimum X value, which can be the left coordinate
        or right coordinate X value.
        """
        return min([self.first_coordinate.x, self.second_coordinate.x])
    
    @property
    def max_y(self):
        """
        The maximum Y value, which can be the left coordinate
        or right coordinate Y value.
        """
        return max([self.first_coordinate.y, self.second_coordinate.y])
    
    @property
    def min_y(self):
        """
        The minimum Y value, which can be the left coordinate
        or right coordinate Y value.
        """
        return min([self.first_coordinate.y, self.second_coordinate.y])
    
    @property
    def is_backwards(self):
        """
        Check if the X value of the first coordinate is
        greater than the X value of the second coordinate.
        If so, the pair of coordinates is backwards (in X
        value, as it is decreasing when going from first
        coordinate to second coordinate).
        """
        return self.first_coordinate.x > self.second_coordinate.x

    @property
    def is_descendant(self):
        """
        Check if the Y value of the first coordinate is
        greater than the Y value of the second coordinate.
        If so, the pair of coordinates is descendant (in
        Y value, as it is decreasing when going from first
        coordinate to second coordinate).
        """
        return self.first_coordinate.y > self.second_coordinate.y

    def __init__(
        self,
        first_coordinate: Coordinate,
        second_coordinate: Coordinate,
        rate_function: RateFunctionArgument = RateFunctionArgument(RateFunction.EASE_IN_OUT_SINE)
    ):
        self.first_coordinate = first_coordinate
        self.second_coordinate = second_coordinate
        self.rate_function = rate_function
    
    def get_xy_from_normalized_local_d(
        self,
        d: float
    ):
        """
        The 'd' parameter must be the local distance within
        this pair of coordinates, that would have been
        calculated from a global map 'd' distance.

        This is the only implemented method because there is
        a floating decimal bug that, when converted to an X
        value, is sometimes out of the limits and raising
        Exception. Using a local 'd' is a better way to make
        sure the X obtained fits the range limit.
        """
        x = self.first_coordinate.x + d * (self.second_coordinate.x - self.first_coordinate.x)
        
        return (
            x,
            self._get_y_from_x(x, is_x_normalized = False).value
        )
    
    def get_n_xy_values_to_plot(
        self,
        n: int = 100,
        do_normalize: bool = False
    ) -> list[tuple[NormalizableValue, NormalizableValue]]:
        """
        Return 'n' (x, y) values to be plotted. Each of those
        X and Y values are normalized only if 'do_normalize'
        flag parameter is set as True.
        """
        if not NumberValidator.is_positive_number(n):
            raise Exception('The provided "n" parameter is not a positive number.')
        
        n = int(n)

        xs = [
            NormalizableValue(x, (self.min_x, self.max_x))
            for x in Progression(self.min_x, self.max_x, 100, RateFunctionArgument.default()).values
        ]
        ys = [
            self._get_y_from_x(x.value, is_x_normalized = False)
            for x in xs
        ]

        if do_normalize:
            xs = [x.normalized for x in xs]
            ys = [y.normalized for y in ys]
        else:
            xs = [x.value for x in xs]
            ys = [y.value for y in ys]

        return list(zip(xs, ys))
    
    def _get_y_from_x(
        self,
        x: float,
        is_x_normalized: bool = False
    ) -> NormalizableValue:
        """
        Get the Y value for the given X, depending on if
        the X value is normalized or not, flagged with the
        'is_x_normalized' parameter.
        
        This method is for internal use only.
        """
        lower_limit = self.min_x if not is_x_normalized else 0
        upper_limit = self.max_x if not is_x_normalized else 1

        # I round the 'x' here to avoid floating decimal issue
        x = round(x)

        if not NumberValidator.is_number_between(x, lower_limit, upper_limit):
            raise Exception(f'The "x" parameter must be between [{lower_limit}, {upper_limit}].')
        
        value = NormalizableValue(x, (self.min_x, self.max_x), value_is_normalized = is_x_normalized)
        # TODO: I think if first x is greater than second x here
        # I must do 1 - self.rate_function.get_n_value(value.normalized)
        # c1(400, 20), c2(300, 10), x = 340
        # 340 normalized is 0.4 [300, 400], but as we are from 400
        # to 300 it should be d=0.6
        value = NormalizableValue(self.rate_function.get_n_value(1 - value.normalized if self.is_backwards else value.normalized), (self.min_y, self.max_y), value_is_normalized = True)
        value = NormalizableValue(1 - value.normalized, (self.min_y, self.max_y), value_is_normalized = True) if self.is_descendant else value

        return value